"""Internal functions for package ssb-hermes!"""

from typing import Any

import pandas as pd  # type: ignore
from thefuzz import process  # type: ignore


def _add_row(
    columns: tuple[str, str, str, str, str],
    *args: str | str | str | Any | str | int,
) -> dict[str, str | str | str | Any | str | int]:
    """Function to add row to dataframe.

    Args:
        columns: Tuple with column names.
        *args: Tuple with values.

    Returns:
        dict: Dictionary with values.
    """
    item = {
        columns[0]: args[0],
        f"{columns[1]}_data": args[1],
        f"{columns[1]}_registry": args[2],
        f"{columns[2]}_data": args[3],
        f"{columns[2]}_registry": args[4],
        "rule_d": args[5],
    }
    return item


def _find_closest_value(
    df: pd.DataFrame,
    column: str,
    value: str,
    score_cutoff: int = 40,
) -> Any:
    """Function to find closest value in column of df.

    Args:
        df: Pandas dataframe containing the data.
        column: String value with name of column in which to look.
        value: String value that we are looking for.
        score_cutoff: Score cutoff. Defaults to 40.

    Returns:
        Any: String with matching item or None.
    """
    choices = df[column].to_list()
    # Har satt cutoff 40 prosent siden det er for gjort å ha 50 % feil med fire siffer
    item = process.extractOne(query=value, choices=choices, score_cutoff=score_cutoff)

    if item is None:
        return None
    else:
        return item[0]


def _check_for_value(
    df: pd.DataFrame,
    column: str,
    value: str,
) -> bool:
    """Function to check for value in column of df.

    Args:
        df: Pandas dataframe containing the data.
        column: String value with name of column in which to look.
        value: String value that we are looking for.

    Returns:
        bool: True or False.
    """
    if value in df[column].to_numpy():
        return True
    else:
        return False


def _check_all_values_equal(
    df: pd.DataFrame,
    column: str,
) -> bool:
    """Function to check if values in column of df are all equal.

    Args:
        df: Pandas dataframe containing the data.
        column: String value with name of column in which to look.

    Returns:
        bool: True or False.
    """
    if df[column].nunique() == 1:
        return True
    else:
        return False


def _get_value_from_df(
    df: pd.DataFrame,
    column1: str,
    column2: str,
    item: str,
) -> Any:
    """Function to get value from df.

    Args:
        df: Pandas dataframe containing the data.
        column1: String value with name of column in which to filter rows.
        column2: String value with name of column in which to get value from.
        item: String value that we are filtering on.

    Returns:
        Any: String value or None
    """
    filtered_rows = df[df[column1] == item]

    # Retrieve values from a specific column in the filtered rows
    values_from_filtered_rows = filtered_rows[column2].unique()[0]
    return values_from_filtered_rows


def _create_list_df_unique_value(
    df: pd.DataFrame,
    column_list: str,
    column_match: str,
    value: str,
) -> Any:
    """Function to create list from df with unique values.

    Args:
        df: Pandas dataframe containing the data.
        column_list: Column to create list from.
        column_match: String value with name of column in which to filter rows.
        value: String value that we are filtering on.

    Returns:
        Any: List with values.
    """
    return (df.loc[df[column_match] == value, column_list]).to_list()


def _set_score_cutoff(df_katalog_subset2: pd.DataFrame, column: str) -> int:
    """Function to set score cutoff.

    Args:
        df_katalog_subset2: Pandas dataframe containing the data.
        column: String value with name of column in which to look.

    Returns:
        int: Integer value.
    """
    if _check_all_values_equal(df_katalog_subset2, column):
        score_cutoff = 0
    else:
        score_cutoff = 75

    return score_cutoff


def _find_postnr_through_adress(
    df_registry_subset: pd.DataFrame,
    liste_data: list[str],
    postnr: str,
    column1: str,
    column2: str,
) -> Any:
    """Function to find postnr through adress.

    Args:
        df_registry_subset: Pandas dataframe containing the data.
        liste_data: List with values.
        postnr: String value with postnr.
        column1: String with column name.
        column2: String with column name.

    Returns:
        Any: String value.
    """
    item = None

    for adresse in liste_data:
        item = _find_closest_value(
            df_registry_subset, column1, adresse, score_cutoff=75
        )
        if item is None:
            continue

        else:
            item = _get_value_from_df(df_registry_subset, column1, column2, item)
            # Om det finnes en match. Da stopper vi loopen og lager en liste med matchen.
            break
    else:
        # Dersom vi når enden av adresene i ibk og ikke har en match, gir vi opp og går videre til neste postnr.
        item = _find_closest_value(df_registry_subset, column2, postnr, score_cutoff=50)

    return item



"""Main function for adress matching.

The template and this example uses Google style docstrings as described at:
https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html

"""

"""Importing packages"""
import os
from typing import Any

import pandas as pd  # type: ignore

"""Importing other internal functions"""
from ._find_match_rules import _find_match_rule1
from ._find_match_rules import _find_match_rule2
from ._find_match_rules import _find_match_rule3
from ._functions import _add_row
from ._functions import _check_all_values_equal
from ._functions import _check_for_value
from ._functions import _create_list_df_unique_value
from ._functions import _find_postnr_through_adress


def find_match(
    df_data: pd.DataFrame,
    df_registry: pd.DataFrame,
    *columns: str,
    registry_type_columns: str,
    find_postnr: bool = False,
) -> tuple[Any, Any, Any]:
    """Fuction for matching adresses from data to registry.

    Args:
        df_data: Dataframe containing the data to be matched.
        df_registry: Dataframe containing the registry.
        *columns: Columns containing the data to be matched. The first column should be the group.
        registry_type_columns: Columns containing the registry type.
        find_postnr: Boolean value. If True, the function will try to find the postnr through the adress. Default is False.

    Returns:
        tuple: A tuple containing rows to make the matched df from, filtered df with wrong postid, filtered df with fauilty adresses.

    Raises:
        ValueError: If the number of columns is not 5.
    """
    if len(columns) != 5:
        raise ValueError("Exactly five columns are required.")
    # Teller enheter
    i = 0
    items = []
    cant_find_postnr = []
    row_failed_no_adress = []

    for group in df_data[columns[0]].unique():
        # Subsetter ut gjeldende foretak fra vof og data inn
        df_registry_subset = df_registry.loc[df_registry[columns[0]] == group, :]
        df_data_subset = df_data.loc[df_data[columns[0]] == group, :]

        # Itererer gjennom et og et postnummer.
        for postnr in df_data_subset[columns[2]].unique():
            liste_data = None
            liste_registry = None
            postnr_registry = None

            liste_data = _create_list_df_unique_value(
                df_data_subset, columns[1], columns[2], postnr
            )

            # Sjekker om postnr finns på foretaket i vof
            if _check_for_value(df_registry_subset, columns[2], postnr):
                postnr_registry = postnr
            else:
                if find_postnr is True:
                    item = None

                    item = _find_postnr_through_adress(
                        df_registry_subset,
                        liste_data,
                        postnr,
                        columns[1],
                        columns[2],
                    )

                    if item is not None:
                        postnr_registry = item
                    else:
                        cant_find_postnr.append(
                            df_data_subset[df_data_subset[columns[1]] == postnr]
                        )
                        continue
                else:
                    continue

            liste_registry = _create_list_df_unique_value(
                df_registry_subset, columns[1], columns[2], postnr_registry
            )

            # Itererer gjennom hver adresse vi får inn på gitt orgnr og postnummer
            for adresse in liste_data:
                item = None
                # Sjekker om det kun finnes en enhet på postnummeret i vof. Isåfall bruker jeg denne enheten
                if len(liste_registry) == 1:
                    item, rule = _find_match_rule1(liste_registry)
                # Om det fins flere enheter på samme postnr bruker jeg thefuzz for å matche dem.
                else:
                    if _check_all_values_equal(
                        df_registry_subset, registry_type_columns
                    ):
                        score_cutoff = 50
                    else:
                        score_cutoff = 75
                    item, rule = _find_match_rule2(
                        adresse, liste_registry, score_cutoff=score_cutoff
                    )
                    # Hvis fuzzy matching ikke funker, sjekker jeg om alle enhetene har samme nace. Isåfall setter jeg bare på en enhet som ikke har None som adresse.

                    # Hvis regel 2 ikke funket, bruker jeg regel 3 til å se etter nærmeste enhet.
                    if item is None:
                        item, rule = _find_match_rule3(
                            df_data_subset,
                            df_registry_subset,
                            postnr,
                            adresse,
                            columns,
                        )
                        # Dersom jeg ikke finner noen match, legger jeg til enheten i en liste.
                        if item is None:
                            row_failed_no_adress.append(
                                df_data_subset[df_data_subset[columns[1]] == adresse]
                            )
                            # Hvis adressen ikke finnes fortsetter vi paa neste adresse.
                            continue

                # Store information about ident from data and registry as a dict to be made into a df.
                items.append(
                    _add_row(
                        columns,
                        group,
                        adresse,
                        item,
                        postnr_registry,
                        postnr,
                        rule,
                    )
                )

                i += 1

                if i % 1000 == 0:
                    print(f"Vi har nådd nr {i} av {len(df_data)}")

    print(f"Ferdig å kjøre {i} enhteter, fant {len(items)} enheter.")
    return items, cant_find_postnr, row_failed_no_adress


def get_test_data() -> tuple[pd.DataFrame, pd.DataFrame]:
    """Function for getting test data.

    Returns:
        tuple:Two dataframes containing test data.
    """
    data_folder = os.path.join(os.path.dirname(__file__), "example_data")
    test_data = os.path.join(data_folder, "test_data.csv")
    test_registry = os.path.join(data_folder, "test_data.csv")

    # Read the CSV file using Pandas and return the DataFrame
    df1 = pd.read_csv(
        test_data,
        header=0,
        names=[
            "ident_gruppe",
            "ident_type",
            "ident_adresse",
            "ident_postnr",
            "ident_kommunenr",
            "ident_fylkenr",
        ],
        dtype={
            "ident_gruppe": str,
            "ident_type": str,
            "ident_adresse": str,
            "ident_postnr": str,
            "ident_kommunenr": str,
            "ident_fylkenr": str,
        },
    )
    df2 = pd.read_csv(
        test_registry,
        header=0,
        names=[
            "ident_gruppe",
            "ident_type",
            "ident_adresse",
            "ident_postnr",
            "ident_kommunenr",
            "ident_fylkenr",
        ],
        dtype={
            "ident_gruppe": str,
            "ident_type": str,
            "ident_adresse": str,
            "ident_postnr": str,
            "ident_kommunenr": str,
            "ident_fylkenr": str,
        },
    )
    return df1, df2



"""Function for each rule in adress matching function.

This python file contain the three functions for adress matching based on the three rules:
    * Rule1: If there is only one unit at a location(postnr), then we use this unit.
    * Rule2: If there are multiple units we use thefuzz with 75% match.
    * Rule3: If rule 1 and 2 did not work, we iterate up geographically.
"""

"""Importing packages"""
from typing import Any

import pandas as pd  # type: ignore
from thefuzz import process  # type: ignore

"""Importing other internal functions"""
from ._functions import _check_for_value
from ._functions import _find_closest_value
from ._functions import _get_value_from_df


def _find_match_rule1(list_with_one: list[str]) -> tuple[str, int]:
    """Function for rule 1 in adress matching function. If there is only one unit at a location(postnr), then we use this unit.

    Args:
        list_with_one: List with one unit at a location(postnr).

    Returns:
        tuple: item, rule
    """
    item = list_with_one[0]
    rule = 1
    return item, rule


def _find_match_rule2(
    query: str, choices: list[str], score_cutoff: int = 75
) -> tuple[Any, Any]:
    """Function for rule 2 in adress matching function. If there are multiple units we use fuzzywuzzy with 75% match.

    Args:
        query: The query string to match.
        choices: List of choices to match against.
        score_cutoff: The score to beat in order to be considered a match.

    Returns:
        tuple: item, rule
    """
    try:
        item, _match = process.extractOne(
            query=query, choices=choices, score_cutoff=score_cutoff
        )
    except (ValueError, AttributeError, TypeError):
        item = None
    finally:
        if item is None:
            rule = None
        else:
            rule = 2
    return item, rule


def _find_match_rule3(
    df_data_subset: pd.DataFrame,
    df_registry_subset: pd.DataFrame,
    postnr: str,
    adresse: str,
    columns: tuple[str, str, str, str, str],
) -> tuple[Any, Any]:
    """Function for rule 3 in adress matching function. If rule 1 and 2 did not work, we iterate up geographically.

    Args:
        df_data_subset: Subset of the data dataframe.
        df_registry_subset: Subset of the registry dataframe.
        postnr: The postnr to match against.
        adresse: The adresse to match against.
        columns: The columns to match against.

    Returns:
        tuple: item, rule
    """
    kommunenr = _get_value_from_df(df_data_subset, columns[2], columns[3], postnr)
    # kommunenr = df_data_subset[df_data_subset[columns[2]] == postnr].reset_index().at[0, columns[3]]
    fylkenr = _get_value_from_df(df_data_subset, columns[2], columns[4], postnr)
    # fylkenr = df_data_subset[df_data_subset[columns[2]] == postnr].reset_index().at[0, columns[4]]

    item = None
    rule = None

    if _check_for_value(df_registry_subset, columns[3], kommunenr):
        df_registry_subset2 = df_registry_subset.loc[
            df_registry_subset[columns[3]] == kommunenr
        ]
        item = _find_closest_value(
            df_registry_subset2, columns[1], adresse, score_cutoff=75
        )

    if item is None:
        if _check_for_value(df_registry_subset, columns[4], fylkenr):
            df_registry_subset2 = df_registry_subset.loc[
                df_registry_subset[columns[4]] == fylkenr
            ]
            item = _find_closest_value(
                df_registry_subset2, columns[1], adresse, score_cutoff=75
            )

    if item is not None:
        rule = 3

    return item, rule
